@@ -25,8 +25,6 @@
 #endif
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#include "xf86Resources.h"
-#include "xf86RAC.h"
 #include "xf86cmap.h"
 #include "compiler.h"
 #include "mibstore.h"
@@ -152,6 +150,13 @@ static void ASTSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
 static void ASTSetClippingRectangle(ScrnInfoPtr pScrn,
                                     int left, int top, int right, int bottom);
 static void ASTDisableClipping(ScrnInfoPtr pScrn); 
+static void ASTSetHWClipping(ScrnInfoPtr pScrn, int delta_y);
+
+static void AIPSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
+                                           int x1, int y1, int x2, int y2, int flags);                                       
+static void AIPSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
+                                            int x1, int y1, int x2, int y2,
+                                            int flags, int phase);
 
 Bool
 ASTAccelInit(ScreenPtr pScreen)
@@ -190,17 +195,33 @@ ASTAccelInit(ScreenPtr pScreen)
     /* Solid Lines */
     if (pAST->ENGCaps & ENG_CAP_SolidLine)
     {    
-        infoPtr->SetupForSolidLine = ASTSetupForSolidLine;
+        if (pAST->jChipType == AST2300)
+    	{
+            infoPtr->SubsequentSolidTwoPointLine = AIPSubsequentSolidTwoPointLine;
+        }
+        else
+    	{
+            infoPtr->SubsequentSolidTwoPointLine = ASTSubsequentSolidTwoPointLine;
+        }
+
+        infoPtr->SetupForSolidLine = ASTSetupForSolidLine;            
         infoPtr->SubsequentSolidHorVertLine = ASTSubsequentSolidHorVertLine;    
-        infoPtr->SubsequentSolidTwoPointLine = ASTSubsequentSolidTwoPointLine;
         infoPtr->SolidLineFlags = NO_PLANEMASK;
     }
 
     /* Dashed Lines */
     if (pAST->ENGCaps & ENG_CAP_DashedLine)
-    {        
-        infoPtr->SetupForDashedLine = ASTSetupForDashedLine;
-        infoPtr->SubsequentDashedTwoPointLine = ASTSubsequentDashedTwoPointLine;
+    {
+        if (pAST->jChipType == AST2300)
+        {
+            infoPtr->SubsequentDashedTwoPointLine = AIPSubsequentDashedTwoPointLine;
+        }
+        else
+        {
+            infoPtr->SubsequentDashedTwoPointLine = ASTSubsequentDashedTwoPointLine;
+        }        
+
+        infoPtr->SetupForDashedLine = ASTSetupForDashedLine;        
         infoPtr->DashPatternMaxLength = 64;
         infoPtr->DashedLineFlags = NO_PLANEMASK |
 			           LINE_PATTERN_MSBFIRST_LSBJUSTIFIED;
@@ -338,6 +359,7 @@ ASTSubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2,
     PKT_SC *pSingleCMD;
     int src_x, src_y, dst_x, dst_y;
     ULONG srcbase, dstbase, cmdreg;
+    int delta_y = 0;
 /*
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentScreenToScreenCopy\n");
 */
@@ -351,22 +373,30 @@ ASTSubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2,
         else
             cmdreg &= ~CMD_ENABLE_CLIP;        
         srcbase = dstbase = 0;
+
+        if (x1 < x2)
+            cmdreg |= CMD_X_DEC;     	
+
+        if (y1 < y2)
+            cmdreg |= CMD_Y_DEC;     	
     
-        if (y1 >= MAX_SRC_Y)
+        if ((y1 + height) >= MAX_SRC_Y)
         {       
             srcbase=pAST->VideoModeInfo.ScreenPitch*y1;
+            y1 = 0;
         }       
         
-        if (y2 >= pScrn->virtualY) 
-        {   
+        if ((y2 + height) >= pScrn->virtualY) 
+        {  
+            delta_y = y2;
             dstbase=pAST->VideoModeInfo.ScreenPitch*y2;
+            y2 = 0;
         }
               
-        if (x1 < x2)
+        if (cmdreg & CMD_X_DEC)
         {
-        	src_x = x1 + width - 1;
-        	dst_x = x2 + width - 1;
-        	cmdreg |= CMD_X_DEC;     	
+            src_x = x1 + width - 1;
+            dst_x = x2 + width - 1;
         }
         else
         {
@@ -374,23 +404,19 @@ ASTSubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int x1, int y1, int x2,
             dst_x = x2;	
         }
      
-        if (y1 < y2)
-        {
-            if (srcbase) y1 = 0;
-            if (dstbase) y2 = 0;
-        	
+        if (cmdreg & CMD_Y_DEC)
+        {        	
             src_y = y1 + height - 1;
             dst_y = y2 + height - 1;
-            cmdreg |= CMD_Y_DEC;     	
         }
         else
         {
-            if (srcbase) y1 = 0;
-            if (dstbase) y2 = 0;
-        	
             src_y = y1;
             dst_y = y2;	
         }
+        
+        if (pAST->EnableClip)                    
+            ASTSetHWClipping(pScrn, delta_y);
             
         if (!pAST->MMIO2D)        
         {
@@ -488,7 +514,9 @@ ASTSubsequentSolidFillRect(ScrnInfoPtr pScrn,
 {
     ASTRecPtr pAST = ASTPTR(pScrn);
     PKT_SC *pSingleCMD;
-    ULONG dstbase, cmdreg;        
+    ULONG dstbase, cmdreg; 
+    int delta_y = 0;
+           
 /*            
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentSolidFillRect\n");
 */
@@ -504,11 +532,15 @@ ASTSubsequentSolidFillRect(ScrnInfoPtr pScrn,
         dstbase = 0;
         
         if (dst_y >= pScrn->virtualY) 
-        {   
+        {
+            delta_y = dst_y;	   
             dstbase=pAST->VideoModeInfo.ScreenPitch*dst_y;
             dst_y=0;
         }
-                      
+
+        if (pAST->EnableClip)                    
+            ASTSetHWClipping(pScrn, delta_y);
+                                  
         if (!pAST->MMIO2D)                    
         {                  
             /* Write to CMDQ */    
@@ -606,6 +638,7 @@ static void ASTSubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
     PKT_SC *pSingleCMD;
     ULONG dstbase, cmdreg;   
     int width, height;
+    int delta_y = 0;
 /*                    
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentSolidHorVertLine\n");
 */
@@ -629,12 +662,15 @@ static void ASTSubsequentSolidHorVertLine(ScrnInfoPtr pScrn,
         }
                   
         if ((y + height) >= pScrn->virtualY) 
-        {   
+        { 
+            delta_y = y;	  
             dstbase=pAST->VideoModeInfo.ScreenPitch*y;
             y=0;
         }
                   
-        
+        if (pAST->EnableClip)                    
+            ASTSetHWClipping(pScrn, delta_y);
+                                          
         if (!pAST->MMIO2D)                    
         {                  
             /* Write to CMDQ */    
@@ -678,6 +714,8 @@ static void ASTSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
     ULONG 	dstbase, ulCommand;
     ULONG	miny, maxy;         
     USHORT      usXM;
+    int delta_y = 0;
+    
 /*
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentSolidTwoPointLine\n");
 */    
@@ -696,6 +734,7 @@ static void ASTSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
     miny = (y1 > y2) ? y2 : y1;
     maxy = (y1 > y2) ? y1 : y2;
     if(maxy >= pScrn->virtualY) {
+    	delta_y = miny;
         dstbase = pAST->VideoModeInfo.ScreenPitch * miny;
         y1 -= miny;
         y2 -= miny;
@@ -714,6 +753,9 @@ static void ASTSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
         ulCommand |= CMD_Y_DEC;             
         
     usXM = (dsLineParam.dwLineAttributes & LINEPARAM_XM) ? 1:0;    
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
        
     if (!pAST->MMIO2D)                    
     {                  
@@ -841,6 +883,8 @@ ASTSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
     ULONG 	dstbase, ulCommand; 
     ULONG	miny, maxy;  
     USHORT      usXM;
+    int delta_y = 0;
+    
 /*
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentDashedTwoPointLine\n");
 */   
@@ -859,6 +903,7 @@ ASTSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
     miny = (y1 > y2) ? y2 : y1;
     maxy = (y1 > y2) ? y1 : y2;
     if(maxy >= pScrn->virtualY) {
+    	delta_y = miny;
         dstbase = pAST->VideoModeInfo.ScreenPitch * miny;
         y1 -= miny;
         y2 -= miny;
@@ -877,6 +922,9 @@ ASTSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
         ulCommand |= CMD_Y_DEC;             
         
     usXM = (dsLineParam.dwLineAttributes & LINEPARAM_XM) ? 1:0;    
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
        
     if (!pAST->MMIO2D)                    
     {                  
@@ -991,7 +1039,9 @@ ASTSubsequentMonoPatternFill(ScrnInfoPtr pScrn,
 {
     ASTRecPtr pAST = ASTPTR(pScrn);
     PKT_SC *pSingleCMD;
-    ULONG dstbase, cmdreg;        
+    ULONG dstbase, cmdreg;
+    int delta_y = 0;
+            
 /*            
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentMonoPatternFill\n");
 */    
@@ -1006,10 +1056,14 @@ ASTSubsequentMonoPatternFill(ScrnInfoPtr pScrn,
 
     if (dst_y >= pScrn->virtualY) 
     {   
+    	delta_y = dst_y;
         dstbase=pAST->VideoModeInfo.ScreenPitch*dst_y;
         dst_y=0;
     }
-                  
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
+                         
     if (!pAST->MMIO2D)                    
     {                  
         /* Write to CMDQ */    
@@ -1116,7 +1170,9 @@ ASTSubsequentColor8x8PatternFillRect(ScrnInfoPtr pScrn, int patx, int paty,
 {
     ASTRecPtr pAST = ASTPTR(pScrn);
     PKT_SC *pSingleCMD;
-    ULONG dstbase, cmdreg;        
+    ULONG dstbase, cmdreg;
+    int delta_y = 0;
+            
 /*            
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentColor8x8PatternFillRect\n");
 */
@@ -1131,9 +1187,13 @@ ASTSubsequentColor8x8PatternFillRect(ScrnInfoPtr pScrn, int patx, int paty,
 
     if (dst_y >= pScrn->virtualY) 
     {   
+    	delta_y = dst_y;
         dstbase=pAST->VideoModeInfo.ScreenPitch*dst_y;
         dst_y=0;
     }
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
                   
     if (!pAST->MMIO2D)                    
     {                  
@@ -1235,6 +1295,7 @@ ASTSubsequentCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
     ASTRecPtr pAST = ASTPTR(pScrn);
     PKT_SC *pSingleCMD;
     ULONG dstbase, cmdreg;
+    int delta_y = 0;
 
 /*           
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentCPUToScreenColorExpandFill\n");
@@ -1250,10 +1311,14 @@ ASTSubsequentCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
 
     if (dst_y >= pScrn->virtualY) 
     {   
+    	delta_y = dst_y;
         dstbase=pAST->VideoModeInfo.ScreenPitch*dst_y;
         dst_y=0;
     }
-                  
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
+                                    
     if (!pAST->MMIO2D)                    
     {                  
         /* Write to CMDQ */    
@@ -1364,6 +1429,7 @@ ASTSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
     PKT_SC *pSingleCMD;
     ULONG srcbase, dstbase, cmdreg;
     USHORT srcpitch;
+    int delta_y = 0;
 
 /*           
     xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentScreenToScreenColorExpandFill\n");
@@ -1378,11 +1444,15 @@ ASTSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
     dstbase = 0;
     if (dst_y >= pScrn->virtualY) 
     {   
+    	delta_y = dst_y;
         dstbase=pAST->VideoModeInfo.ScreenPitch*dst_y;
         dst_y=0;
     }
     srcbase = pAST->VideoModeInfo.ScreenPitch*src_y + ((pScrn->bitsPerPixel+1)/8)*src_x;            
     srcpitch = (pScrn->displayWidth+7)/8;
+
+    if (pAST->EnableClip)                    
+        ASTSetHWClipping(pScrn, delta_y);
     
     if (!pAST->MMIO2D)                    
     {                  
@@ -1422,6 +1492,33 @@ ASTSubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
 
 	  
 /* Clipping */
+static void
+ASTSetHWClipping(ScrnInfoPtr pScrn, int delta_y)
+{
+    ASTRecPtr pAST = ASTPTR(pScrn);
+    PKT_SC *pSingleCMD;
+
+    if (!pAST->MMIO2D)                    
+    {
+        /* Write to CMDQ */
+        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pAST, PKT_SINGLE_LENGTH*2);
+
+        ASTSetupCLIP1(pSingleCMD, pAST->clip_left, pAST->clip_top - delta_y);
+        pSingleCMD++;
+        ASTSetupCLIP2(pSingleCMD, pAST->clip_right + 1, pAST->clip_bottom - delta_y + 1);
+        
+        /* Update Write Pointer */
+        mUpdateWritePointer;
+                         
+    }
+    else
+    {
+        ASTSetupCLIP1_MMIO(pAST->clip_left, pAST->clip_top - delta_y);
+        ASTSetupCLIP2_MMIO(pAST->clip_right + 1, pAST->clip_bottom - delta_y + 1);                   	
+    }
+	
+}	
+
 static void
 ASTSetClippingRectangle(ScrnInfoPtr pScrn,
                         int left, int top, int right, int bottom)
@@ -1434,24 +1531,10 @@ ASTSetClippingRectangle(ScrnInfoPtr pScrn,
 */            
     pAST->EnableClip = TRUE;
             
-    if (!pAST->MMIO2D)                    
-    {
-        /* Write to CMDQ */
-        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pAST, PKT_SINGLE_LENGTH*2);
-
-        ASTSetupCLIP1(pSingleCMD, left, top);
-        pSingleCMD++;
-        ASTSetupCLIP2(pSingleCMD, right, bottom);
-        
-        /* Update Write Pointer */
-        mUpdateWritePointer;
-                         
-    }
-    else
-    {
-        ASTSetupCLIP1_MMIO(left, top);
-        ASTSetupCLIP2_MMIO(right, bottom);                   	
-    }
+    pAST->clip_left   = left;
+    pAST->clip_top    = top;
+    pAST->clip_right  = right;
+    pAST->clip_bottom = bottom;
     
 }
 
@@ -1465,4 +1548,141 @@ ASTDisableClipping(ScrnInfoPtr pScrn)
     pAST->EnableClip = FALSE;
 }
 
+static void AIPSubsequentSolidTwoPointLine(ScrnInfoPtr pScrn,
+                                           int x1, int y1, int x2, int y2, int flags)
+{
+ 
+    ASTRecPtr 	pAST = ASTPTR(pScrn);
+    PKT_SC 	*pSingleCMD;
+    ULONG 	dstbase, ulCommand;
+    ULONG	miny, maxy;         
+/*
+    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentSolidTwoPointLine\n");
+*/    
+
+    /* Modify Reg. Value */
+    ulCommand = (pAST->ulCMDReg & (~CMD_MASK)) | CMD_LINEDRAW | CMD_NORMAL_LINE;
+    if(flags & OMIT_LAST)
+        ulCommand |= CMD_NOT_DRAW_LAST_PIXEL;
+    else
+        ulCommand &= ~CMD_NOT_DRAW_LAST_PIXEL;
+    if (pAST->EnableClip)
+        ulCommand |= CMD_ENABLE_CLIP;
+    else
+        ulCommand &= ~CMD_ENABLE_CLIP;        
+    dstbase = 0;
+    miny = (y1 > y2) ? y2 : y1;
+    maxy = (y1 > y2) ? y1 : y2;
+    if(maxy >= pScrn->virtualY) {
+        dstbase = pAST->VideoModeInfo.ScreenPitch * miny;
+        y1 -= miny;
+        y2 -= miny;
+    }
+
+    if (!pAST->MMIO2D)                    
+    {                  
+        /* Write to CMDQ */    
+        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pAST, PKT_SINGLE_LENGTH*5);
+
+        ASTSetupDSTBase(pSingleCMD, dstbase);
+        pSingleCMD++;    
+        AIPSetupLineXY(pSingleCMD, x1, y1);
+        pSingleCMD++;
+        AIPSetupLineXY2(pSingleCMD, x2, y2);
+        pSingleCMD++;
+        AIPSetupLineNumber(pSingleCMD, 0);
+        pSingleCMD++;                     
+        ASTSetupCMDReg(pSingleCMD, ulCommand);        
+              
+        /* Update Write Pointer */
+        mUpdateWritePointer;                
+
+        /* Patch KDE pass abnormal point, ycchen@052507 */
+        vWaitEngIdle(pScrn, pAST);
+        
+    }
+    else
+    {                  
+        ASTSetupDSTBase_MMIO(dstbase);
+        AIPSetupLineXY_MMIO(x1, y1);
+        AIPSetupLineXY2_MMIO(x2, y2);
+        AIPSetupLineNumber_MMIO(0);
+        ASTSetupCMDReg_MMIO(ulCommand);        
+             
+        vWaitEngIdle(pScrn, pAST);
+      
+    }
+
+                
+} /* end of AIPSubsequentSolidTwoPointLine */
+
+static void
+AIPSubsequentDashedTwoPointLine(ScrnInfoPtr pScrn,
+                                int x1, int y1, int x2, int y2,
+                                int flags, int phase)
+{
+ 
+    ASTRecPtr 	pAST = ASTPTR(pScrn);
+    PKT_SC 	*pSingleCMD;
+    ULONG 	dstbase, ulCommand; 
+    ULONG	miny, maxy;  
+/*
+    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ASTSubsequentDashedTwoPointLine\n");
+*/   
+
+    /* Modify Reg. Value */
+    ulCommand = pAST->ulCMDReg | CMD_NORMAL_LINE;
+    if(flags & OMIT_LAST)
+        ulCommand |= CMD_NOT_DRAW_LAST_PIXEL;
+    else
+        ulCommand &= ~CMD_NOT_DRAW_LAST_PIXEL;
+    if (pAST->EnableClip)
+        ulCommand |= CMD_ENABLE_CLIP;
+    else
+        ulCommand &= ~CMD_ENABLE_CLIP;        
+    dstbase = 0;        
+    miny = (y1 > y2) ? y2 : y1;
+    maxy = (y1 > y2) ? y1 : y2;
+    if(maxy >= pScrn->virtualY) {
+        dstbase = pAST->VideoModeInfo.ScreenPitch * miny;
+        y1 -= miny;
+        y2 -= miny;
+    }
+
+    if (!pAST->MMIO2D)                    
+    {                  
+        /* Write to CMDQ */    
+        pSingleCMD = (PKT_SC *) pjRequestCMDQ(pAST, PKT_SINGLE_LENGTH*5);
+
+        ASTSetupDSTBase(pSingleCMD, dstbase);
+        pSingleCMD++;    
+        AIPSetupLineXY(pSingleCMD, x1, y1);
+        pSingleCMD++;
+        AIPSetupLineXY2(pSingleCMD, x2, y2);
+        pSingleCMD++;
+        AIPSetupLineNumber(pSingleCMD, 0);
+        pSingleCMD++;                     
+        ASTSetupCMDReg(pSingleCMD, ulCommand);        
+              
+        /* Update Write Pointer */
+        mUpdateWritePointer;
+
+        /* Patch KDE pass abnormal point, ycchen@052507 */
+        vWaitEngIdle(pScrn, pAST);
+              
+    }
+    else
+    {                  
+        ASTSetupDSTBase_MMIO(dstbase);
+        AIPSetupLineXY_MMIO(x1, y1);
+        AIPSetupLineXY2_MMIO(x2, y2);
+        AIPSetupLineNumber_MMIO(0);
+        ASTSetupCMDReg_MMIO(ulCommand);
+       
+        vWaitEngIdle(pScrn, pAST);
+      
+    }
+                	
+}
+
 #endif	/* end of Accel_2D */
